#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
"""
 Creation of environmental maps

 @author East Japan Accounting Center Co., Ltd

 (c)2022 The University of Aizu
 This software is released under the New BSD License.
"""
import cv2
import numpy as np
import logging
import os
import random
import string
from mongodb import MongoDB
from postgres import PostgreSQL
from env_map_info_notice import json_file_parse
from datetime import datetime, timedelta, timezone
from map_type_info import MapLayerTable

logger = logging.getLogger(__name__)
JST = timezone(timedelta(hours=+9), "JST")
SOCKET_CONF_PATH = "./conf/socket_conf.json"
DB_CONF_PATH = "./conf/db_conf.json"
MAP_STORAGE_CONF_PATH = "./conf/map_storage_conf.json"


class EnvMapBuilder:
    # class variable
    FIXED_MAP_SPACE = "real"
    FIXED_ROBOT_ID = "rdr_01"
    OBJECT_ORIGIN_X = 0.596116
    OBJECT_ORIGIN_Y = -5.97717
    UNOBSERVED_AREA = 255
    NO_OBSTACLE_AREA = 0
    OBSTACLE_AREA = 100
    UNOBSERVED_AREA_COLOR = [127, 127, 127]
    NO_OBSTACLE_AREA_COLOR = [255, 255, 255]
    OBSTACLE_AREA_COLOR = [0, 0, 0]
    NO_OBSTACLE_AREA_RANGE = 170
    OBSTACLE_AREA_RANGE = 85
    BGR_WHITE = (255, 255, 255)
    BGR_BLACK = (0, 0, 0)
    SERIAL_NUMBER_MAX = 100
    STATIC_OBJECT_TABLE_NAME = "Table"
    STATIC_OBJECT_CHAIR_NAME = "chair"

    def __init__(self, _, __, ___, layout_datetime=""):
        self.cnv_coord_param = {"rotation": -90,
                                "correction": (0, 0),
                                "resolution": 0}
        self.resolution = 0
        self.origin_x = 0
        self.origin_y = 0
        self.layout_datetime = layout_datetime
        self.mongodb_conf = None
        self.postgres_conf = None
        self.map_storage_conf = None
        self.mongodb = None
        self.postgres = None
        self.ref_map_info = None
        self.ref_map_width = None
        self.ref_map_height = None
        self.ref_map_map_path = None

    def config_setting(self):
        """
        preparation of configuration data

        Return
        ----------
        result: end of preparation
        """
        result = False
        parse_res, db_conf = json_file_parse(DB_CONF_PATH)
        storage_parse_res, map_storage_conf = json_file_parse(MAP_STORAGE_CONF_PATH)

        if parse_res and storage_parse_res:
            self.mongodb_conf = db_conf["mongodb"]
            self.postgres_conf = db_conf["postgresql"]
            self.map_storage_conf = map_storage_conf
            self.mongodb = MongoDB(self.mongodb_conf["host"], self.mongodb_conf["port"], self.mongodb_conf["db_name"])
            self.postgres = PostgreSQL(self.postgres_conf["db_name"], self.postgres_conf["host"],
                                       self.postgres_conf["tbl_name"], self.postgres_conf["user"],
                                       self.postgres_conf["password"])
            result = True

        return result

    def conv_map_coord(self, x_coord, y_coord):
        """
        convert the position information calculated by the object recognition system into map coordinates

        Param
        ----------
        x_coord: Coordinate information x calculated by the object recognition system Unit (m)
        y_coord: Coordinate information y calculated by the object recognition system Unit (m)

        Return
        ----------
        map_coord: map coordinate data (x, y) of the object converted to pixel coordinates
        """
        # Convert coordinates from [m] to [px]
        map_coord_x = round((x_coord - self.origin_x) / self.resolution)
        map_coord_y = round((y_coord - self.origin_y) / self.resolution)
        map_coord = (int(map_coord_x), int(map_coord_y))
        return map_coord

    def object_2_map_coord(self, object_coord):
        """
        convert the position information calculated by the object recognition system into map coordinates

        Param
        ----------
        object_coord: Position information calculated by the object recognition system (x, y) Unit (m)

        Return
        ----------
        map_coord: map coordinate data (x, y) of the object converted to pixel coordinates
        """
        # Convert from degrees to radians
        rad = np.deg2rad(self.cnv_coord_param["rotation"])
        # Calculate Rotation Matrix
        r_matrix = np.array([[np.cos(rad), -np.sin(rad)], [np.sin(rad), np.cos(rad)]])

        # Distance between origin coordinates on Map side and origin coordinates on external sensor side [m]
        corr_x = self.cnv_coord_param["correction"][0]
        corr_y = self.cnv_coord_param["correction"][1]
        # Align the coordinate axes of the map side and the external sensor side
        calc_mat = np.dot(r_matrix, object_coord)

        # Convert coordinates from [m] to [px]
        map_coord_x = round(round((calc_mat[0] + corr_x), 6) / self.cnv_coord_param["resolution"])
        map_coord_y = round(round((calc_mat[1] + corr_y), 6) / self.cnv_coord_param["resolution"])
        map_coord = (int(map_coord_x), int(map_coord_y))
        return map_coord

    # mongodb related methods
    def connect_mongodb(self):
        """
        connect mongodb

        Return
        ----------
        result: connection result
        """
        result = False
        host = self.mongodb_conf["host"]
        port = self.mongodb_conf["port"]
        conn = self.mongodb.connection(host, port)
        if conn is not None:
            result = True
        else:
            logger.error("Unable to connect to MongoDB.")
        return result

    def disconnect_mongodb(self):
        """
        disconnect mongodb
        """
        self.mongodb.close()

    def find_map_mongodb(self, query):
        """
        retrieve map layer from DB

        Param
        ----------
        query: Map search condition query

        Return
        ----------
        doc: search results
        """
        doc = self.mongodb.get_document(self.mongodb_conf["db_name"], self.mongodb_conf["map_collection"], query)
        return doc

    def find_layout_mongodb(self, query):
        """
        retrieve map layer from DB

        Param
        ----------
        query: layout search condition query

        Return
        ----------
        doc: search results
        """
        doc = self.mongodb.get_document(self.mongodb_conf["db_name"], self.mongodb_conf["layout_collection"], query)
        return doc

    def read_reference_map(self):
        """
        Read reference map information and store various data in variables

        Returns
        -------
        result: Result of loading the reference map
        """
        result = False
        # Search layer map matching specified revision, search latest layer map if not found
        map_layer_info = self.find_map_mongodb({"map_layer": MapLayerTable.DICT[MapLayerTable.KEY.REFERENCE]})
        if map_layer_info is None:
            return result

        self.ref_map_info = map_layer_info
        self.ref_map_width = map_layer_info["map_data"]["info"]["width"]
        self.ref_map_height = map_layer_info["map_data"]["info"]["height"]
        self.ref_map_map_path = map_layer_info["map_data"]["data_path"]
        self.resolution = map_layer_info["map_data"]["info"]["resolution"]
        self.origin_x = map_layer_info["map_data"]["info"]["origin"]["position"]["x"]
        self.origin_y = map_layer_info["map_data"]["info"]["origin"]["position"]["y"]
        result = True
        return result

    def get_new_layer_revision(self, layer_type):
        """
        provides new revision numbers for layer maps

        Param
        ----------
        layer_type: specified layer map type

        Return
        ----------
        new_revision: new revision number
        """
        new_revision = None
        map_layer_info = self.find_map_mongodb({"map_layer": layer_type})
        if map_layer_info is None:
            return new_revision
        latest_revision = map_layer_info["revision"]
        # split revision
        revision_a = latest_revision.split("-")
        # add number
        revision_num = int(revision_a[4]) + 1
        # Reshape Revision
        new_revision = "{}-{}-{}-{}-{}".format(
            revision_a[0], revision_a[1], revision_a[2], revision_a[3], str(revision_num))
        return new_revision

    # postgresql related methods
    def insert_env_map_psql(self, param):
        """
        add new environment map data to DB

        Param
        ----------
        param: parameter information of the environment map to be registered

        Return
        ----------
        result: additional result
        """
        result = False
        try:
            with self.postgres.get_connection() as conn:
                # Enable auto-commit
                conn.autocommit = True
                with self.postgres.get_acquisition_cursor(conn) as cur:
                    sql = "INSERT INTO {} (robot_id, type, space, \"time\", revision, path, "\
                          "location, lat, lon, azimuth, resolution, width, height, " \
                          "origin_x, origin_y, origin_z, " \
                          "orientation_x, orientation_y, orientation_z, orientation_w, " \
                          "static_layer_revision, semi_static_layer_revision, " \
                          "exclusion_zone_layer_revision, layout_object_id) " \
                          "VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s ,%s, %s, %s, %s, %s, %s, %s, %s, %s)"
                    # insert environment map information to DB
                    insert_res = self.postgres.insert(cur, sql, param)
                    if insert_res:
                        result = True
        except Exception as e:
            logger.error(e)
        return result

    def get_new_env_revision(self):
        """
        provide a new revision of the environment map

        Return
        ----------
        new_revision: new revision number
        """
        new_revision = None
        try:
            with self.postgres.get_connection() as conn:
                with self.postgres.get_acquisition_cursor(conn) as cur:
                    # SQL to find max revision of environment map
                    request_sql = "SELECT (left(max(revision), 20) || " \
                                  "cast(max(cast(substring(revision, 21, length(revision)) as integer)) " \
                                  "as character varying)) as revision " \
                                  "FROM {} WHERE revision LIKE  {}"
                    param = "'____-____-____-____-%'"
                    select_res, select_data = self.postgres.select(cur, request_sql, param)
                    if select_res:
                        if select_data and select_data[0]["revision"] is not None:
                            # split the fetched revision
                            latest_revision = select_data[0]["revision"]
                            revision_a = latest_revision.split("-")
                            # add revision
                            revision_num = int(revision_a[4]) + 1
                            # Recombine revisions
                            new_revision = "{}-{}-{}-{}-{}".format(revision_a[0], revision_a[1], revision_a[2],
                                                                   revision_a[3], str(revision_num))
                        else:
                            # Create a new revision if no revision is found
                            new_revision = self.create_revision()
        except Exception as e:
            logger.error(e)
        return new_revision

    # File operation related methods
    @staticmethod
    def map_read(file_path):
        """
        provides loaded map file data

        Param
        ----------
        file_path: target map file path

        Return
        ----------
        read_map_data: read map data
        """
        read_map_data = []
        try:
            with open(file_path, "rb") as map_bin:
                map_byte_a = bytearray(map_bin.read())
                read_map_data = np.array(list(map_byte_a))
        except IOError as e:
            logger.error(e)
            logger.error("Unable to access map data.")
        return read_map_data

    def conv_map_2_image(self, map_data_a, width, height):
        """
        Read map data and provide as image data

        Param
        ----------
        map_data_a: Byte type map data
        width: map width size
        height: map height size

        Return
        ----------
        2D array image data
        """
        image_a = None
        data_size = height * width
        if data_size != len(map_data_a):
            return image_a
        # Convert from 1D array to 2D array
        map_data_a = map_data_a.reshape(height, width)
        # create an empty image data array
        image_a = np.zeros((height, width, 3))
        for px_h in range(height):
            for px_w in range(width):
                # Convert from map data value to BGR
                image_a[px_h][px_w] = self.NO_OBSTACLE_AREA_COLOR if map_data_a[px_h][px_w] == self.NO_OBSTACLE_AREA \
                    else self.OBSTACLE_AREA_COLOR if map_data_a[px_h][px_w] == self.OBSTACLE_AREA else self.UNOBSERVED_AREA_COLOR
        image_a = image_a.astype("uint8")
        image_a = cv2.cvtColor(image_a, cv2.COLOR_BGR2GRAY)
        return image_a

    @staticmethod
    def map_write(map_data, file_path):
        """
        write map data to file

        Param
        ----------
        map_data: map sequence data
        file_path: write destination

        Return
        ----------
        result :result written to file
        """
        result = False
        byte_a = bytearray(map_data.astype("uint8").ravel())
        try:
            with open(file_path, "xb") as file:
                file.write(byte_a)
                result = True
        except IOError as e:
            logger.error(e)
            logger.error("Unable to build map data.")
        return result

    def conv_image_2_map_write(self, tgt_image, file_path):
        """
        convert image data to map format and generate file

        Param
        ----------
        tgt_image: target image data
        file_path: file write destination

        Return
        ----------
        result: result of file generate
        """
        result = False
        byte_a = bytearray([])
        for image_array in tgt_image:
            for image_data in image_array:
                # Convert image data to map data
                image_data = self.OBSTACLE_AREA if image_data <= self.OBSTACLE_AREA_RANGE else \
                    self.NO_OBSTACLE_AREA if image_data >= self.NO_OBSTACLE_AREA_RANGE else self.UNOBSERVED_AREA
                byte_a.append(image_data)
        try:
            # Write the converted map data to a file
            with open(file_path, "xb") as file:
                file.write(byte_a)
                result = True
        except OSError as e:
            logger.error(e)
            logger.error("Unable to build map data.")
        return result

    def create_map_file(self, map_image=None, map_array=None, map_layer_type=None):
        """
        create map file data

        Param
        ----------
        map_image: map data created with images
        map_array: binary data type map data
        map_layer_type: create map layer type

        Return
        ----------
        Relative path of map data
        """
        result = None
        # image data
        if map_image is not None:
            create_function = self.conv_image_2_map_write
            set_data = map_image
        # byte array data
        elif map_array is not None:
            create_function = self.map_write
            set_data = map_array
        else:
            return result
        data_path = self.create_map_file_path(map_layer_type)
        if data_path is None:
            return result
        if create_function(set_data, data_path):
            result = data_path.replace(self.map_storage_conf["root"], "")
        return result

    def create_map_file_path(self, layer_type):
        """
        naming layer map file

        Param
        ----------
        layer_type: specifying map layer types

        Return
        ----------
        file_path: path information including map file name
        """
        file_path = None
        if layer_type is None:
            return file_path
        cur_time = datetime.now(JST)
        base_dir = self.map_storage_conf["root"] + self.map_storage_conf["base"]
        time_dir = self.map_storage_conf["time_dir_format"].format(
            year=cur_time.year, month=cur_time.month, day=cur_time.day)
        layer_dir = self.map_storage_conf["map_layer_format"].format(map_layer=layer_type)
        file_name = self.map_storage_conf["file_format"]

        # Define the storage destination directory and generate the directory
        target_dir = base_dir + time_dir + layer_dir
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)

        # Define filename
        for file_num in range(self.SERIAL_NUMBER_MAX):
            set_num = "" if file_num == 0 else str(file_num)
            file_path = target_dir + file_name.format(year=cur_time.year, month=cur_time.month,
                                                      day=cur_time.day, hr=cur_time.hour,
                                                      min=cur_time.minute, sec=cur_time.second, num=set_num)
            if not os.path.isfile(file_path):
                break
        else:
            file_path = None
        return file_path

    @staticmethod
    def create_revision():
        """
        define new revision

        Return
        ----------
        new revision
        """
        new_revision = ""
        for i in range(4):
            random_list = [random.choice(string.ascii_lowercase + string.digits) for _ in range(4)]
            new_revision += "".join(random_list) + "-"
        new_revision += "1"
        return new_revision

    def create_reg_env_map_param(self, semi_st_layer_rev, exc_layer_rev, layout_id, env_map_data_path):
        """
        create parameter information of environment map to be registered in DB
        Param
        ----------
        semi_st_layer_rev: semi static layer map revision info
        exc_layer_rev: exclusion zone layer map revision info
        layout_id: ID information of the layout used for the environment map
        env_map_data_path: map file path

        Return
        ----------
        param: data group registered in DB
        """
        param = None
        base_map_info = self.ref_map_info.copy()
        robot_id = self.FIXED_ROBOT_ID
        map_type = base_map_info["type"]
        space = self.FIXED_MAP_SPACE
        time = "{}+0900".format(datetime.now(JST).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3])
        revision = self.get_new_env_revision()
        if revision is None:
            return param
        location = base_map_info["location_info"]["location"]
        lat = base_map_info["location_info"]["lat"]
        lon = base_map_info["location_info"]["lon"]
        azimuth = base_map_info["location_info"]["azimuth"]
        resolution = self.cnv_coord_param["resolution"]
        origin_x = base_map_info["map_data"]["info"]["origin"]["position"]["x"]
        origin_y = base_map_info["map_data"]["info"]["origin"]["position"]["y"]
        origin_z = base_map_info["map_data"]["info"]["origin"]["position"]["z"]
        orientation_x = base_map_info["map_data"]["info"]["origin"]["orientation"]["x"]
        orientation_y = base_map_info["map_data"]["info"]["origin"]["orientation"]["y"]
        orientation_z = base_map_info["map_data"]["info"]["origin"]["orientation"]["z"]
        orientation_w = base_map_info["map_data"]["info"]["origin"]["orientation"]["w"]
        static_layer_revision = base_map_info["revision"]
        width = self.ref_map_width
        height = self.ref_map_height

        param = (robot_id, map_type, space, time, revision, env_map_data_path,
                 location, lat, lon, azimuth,
                 resolution, width, height, origin_x, origin_y, origin_z,
                 orientation_x, orientation_y, orientation_z, orientation_w,
                 static_layer_revision, semi_st_layer_rev, exc_layer_rev, layout_id,)
        return param

    def register_env_map(self, env_map_data, semi_st_layer_rev, exc_layer_rev, layout_id):
        """
        create data related to the environment map and register it in the DB
        ----------
        st_layer_map_info: static layer map information
        semi_st_layer_rev: semi static layer map revision info
        exc_layer_rev: exclusion zone layer map revision info
        layout_id: ID information of the layout used for the environment map
        width: map width
        height: map height
        env_map_data_path: map file path

        Return
        ----------
        result: registration result to DB
        """
        result = False
        relative_path = self.create_map_file(map_image=env_map_data,
                                             map_layer_type=MapLayerTable.NAME.ENVIRONMENT.value)

        if relative_path is None:
            return result
        pram = self.create_reg_env_map_param(semi_st_layer_rev, exc_layer_rev, str(layout_id), relative_path)
        if pram is None:
            return result
        # Insert environmental map information into DB
        insert_res = self.insert_env_map_psql(pram)
        result = True if insert_res else False
        return result

    def place_layout_on_map(self, map_data_a, map_type, width, height, layout_dict):
        """
        Convert map data to image data and place semi-static objects

        Param
        ----------
        map_data_a: Map image data
        map_type: Map type
        width: Map width
        height: Map height
        layout_dict: Position of the quasi-static object to place

        Returns
        -------
        Map image data with quasi-static objects
        """
        image_data = self.conv_map_2_image(map_data_a, width, height)
        if image_data is None:
            return image_data
        # Place the table on the map
        image_data = self.place_table(layout_dict[self.STATIC_OBJECT_TABLE_NAME], map_type, image_data)
        # Place the chair on the map
        image_data = self.place_chair(layout_dict[self.STATIC_OBJECT_CHAIR_NAME], image_data)
        return image_data

    def register_layer_map(self, image_data, map_type):
        """
        After saving the layer map image, register the layer map information in the DB

        Param
        ----------
        image_data: Map image data
        map_type: Map type

        Return
        ----------
        Returns the revision number if the registration succeeds, or None if it fails
        """
        revision = None
        # Layer map file generation
        relative_path = self.create_map_file(map_image=image_data,
                                             map_layer_type=MapLayerTable.DICT[map_type])
        if relative_path is None:
            return revision
        # Insert updated map information into DB
        set_revision = self.get_new_layer_revision(MapLayerTable.DICT[map_type])
        new_map_info = self.ref_map_info.copy()
        new_map_info.pop("_id", None)
        new_map_info["revision"] = set_revision
        new_map_info["robot_id"] = self.FIXED_ROBOT_ID
        new_map_info["space"] = self.FIXED_MAP_SPACE
        new_map_info["time"] = "{}+0900".format(datetime.now(JST).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3])
        new_map_info["map_layer"] = MapLayerTable.DICT[map_type]
        new_map_info["map_data"]["data_path"] = relative_path

        # Insert layer map to DB
        if self.mongodb.insert_document(self.mongodb_conf["db_name"], self.mongodb_conf["map_collection"], new_map_info):
            revision = set_revision
        return revision

    def place_table(self, table_list_a, map_type, image_data):
        """
        Placing tables on the map image data

        Param
        ----------
        table_list_a: Table arrangement information list
        map_type: Map type
        image_data: Map image data

        Return
        ----------
        Map image data with tables arranged
        """
        table_coord_a = []
        for layout_table in table_list_a:
            # Convert layout coordinates to map coordinates
            table_corners_a = [self.conv_map_coord(corner["x"], corner["y"]) for corner in layout_table["corners"]]
            table_coord_a.append(table_corners_a)
            set_pts = np.array(table_corners_a)

            # Draw table position on image
            if map_type == MapLayerTable.KEY.EXCLUSION_ZONE or map_type == MapLayerTable.KEY.ENVIRONMENT:
                cv2.polylines(image_data, [set_pts], True, color=self.BGR_BLACK, thickness=2)
            if map_type == MapLayerTable.KEY.SEMI_STATIC:
                _ = [cv2.circle(image_data, pts, 3, self.BGR_BLACK, thickness=-1) for pts in set_pts]
        return image_data

    def place_chair(self, chair_list_a, image_data):
        """
        Placing chairs on the map image data

        Param
        ----------
        chair_list_a: Chair arrangement information list
        image_data: Map image data

        Return
        ----------
        Map image data with chairs arranged
        """
        chair_centers_a = [self.conv_map_coord(layout_chair["center"]["x"], layout_chair["center"]["y"])
                           for layout_chair in chair_list_a]
        # Draw chair position on image
        _ = [cv2.circle(image_data, pts, 3, self.BGR_BLACK, thickness=-1) for pts in chair_centers_a]
        return image_data

    def reference_map_data_read(self):
        """
        Read and provide reference maps

        Return
        ----------
        map_data:Reference map data
        width:   Map height
        height:  Map width
        """
        map_data = self.map_read(self.map_storage_conf["root"] + self.ref_map_map_path)
        return map_data, self.ref_map_width, self.ref_map_height


def env_map_builder_proc(option1, option2, option3, option4):
    """
    environment map create process

    Param
    ----------
    option1: static layer map revision number
    option2: semi-static layer map revision number
    option3: static layer map revision number
    option4: layout changed datetime

    Return
    ----------
    result: process result
    """
    result = False
    logger.debug("environment map create process start")
    logger.debug("argument: 1:[{}], 2:[{}], 3:[{}], 4:[{}]".format(option1, option2, option3, option4))
    builder_ins = EnvMapBuilder(option1, option2, option3, option4)
    #            Read config                 Connection DB                Read reference map
    for func in [builder_ins.config_setting, builder_ins.connect_mongodb, builder_ins.read_reference_map]:
        if not func():
            return result

    # Find layout object
    layout_info = builder_ins.find_layout_mongodb({"datetime": builder_ins.layout_datetime})
    if layout_info is None:
        builder_ins.disconnect_mongodb()
        return result
    # Create list of tables from layout
    layout_dict = {builder_ins.STATIC_OBJECT_TABLE_NAME: [layout_coord for layout_coord in layout_info["layout_list"]
                                                          if layout_coord.get("name") == builder_ins.STATIC_OBJECT_TABLE_NAME],
                   builder_ins.STATIC_OBJECT_CHAIR_NAME: [layout_coord for layout_coord in layout_info["layout_list"]
                                                          if layout_coord.get("name") == builder_ins.STATIC_OBJECT_CHAIR_NAME]}

    # Read reference map data
    reference_map_data, width, height = builder_ins.reference_map_data_read()
    if reference_map_data.size == 0:
        builder_ins.disconnect_mongodb()
        return result

    # create base layer map data
    # Replace unknown areas with no obstacle areas
    exc_layer_map_data = np.where(reference_map_data == builder_ins.OBSTACLE_AREA, builder_ins.OBSTACLE_AREA, builder_ins.NO_OBSTACLE_AREA)
    # Creates a layer map of semi-static objects with no obstructions over the entire area
    semi_st_layer_map_data = np.zeros(height * width)

    # Place semi-static objects on the layer map and register the exclusion zone layer map
    exc_layer_map_data = builder_ins.place_layout_on_map(exc_layer_map_data, MapLayerTable.KEY.EXCLUSION_ZONE, width, height, layout_dict)
    exc_layer_revision = builder_ins.register_layer_map(exc_layer_map_data, MapLayerTable.KEY.EXCLUSION_ZONE)

    # Place semi-static objects on the layer map and register the semi-static layer map
    semi_st_layer_map_data = builder_ins.place_layout_on_map(semi_st_layer_map_data, MapLayerTable.KEY.SEMI_STATIC, width, height, layout_dict)
    semi_st_layer_revision = builder_ins.register_layer_map(semi_st_layer_map_data, MapLayerTable.KEY.SEMI_STATIC)

    builder_ins.disconnect_mongodb()
    if exc_layer_revision is None or semi_st_layer_revision is None:
        # Failed to create layer map
        return result

    # Create environment map
    env_map_data = builder_ins.place_layout_on_map(reference_map_data, MapLayerTable.KEY.ENVIRONMENT, width, height, layout_dict)

    # Registration environment map
    register_res = builder_ins.register_env_map(env_map_data, semi_st_layer_revision, exc_layer_revision, layout_info["_id"])
    if register_res:
        # Request map update notifications
        map_update_notice_req()
        result = True
        logger.debug("environment map create process success")
    logger.debug("environment map create process end")
    return result


def map_update_notice_req():
    """
    Request map update notifications
    """
    from rdr_map_manager import ReqMapSocketMsg, ReqMapProcId
    sock_ins = ReqMapSocketMsg(ReqMapProcId.MAP_INFO_NOTICE)
    parse_res, socket_conf = json_file_parse(SOCKET_CONF_PATH)
    if parse_res:
        sock_ins.set_socket_setting(socket_conf["host"], socket_conf["port"])
        sock_ins.send()


if __name__ == "__main__":
    # pass
    env_map_builder_proc("1234-5678-abcd-efgh-1", "1234-5678-abcd-abcd-1",
                         "1234-5678-abcd-1234-1", "2023-08-01T17:33:16.390+0900")
    # ins = EnvMapBuilder("1234-5678-abcd-efgh-1", "1234-5678-abcd-abcd-1",
    #                             "1234-5678-abcd-1234-1", "2022-11-16T08:34:30.133+0900")
    #
    # static_dat = ins.map_read("/var/rdr/web/map/data/static.dat")
    # static_img = ins.conv_map_2_image(static_dat, 749, 531)
    # cv2.imshow("Image", static_img)
    # cv2.waitKey()
    #
    #
    # semi_dat = ins.map_read("/var/rdr/web/map/20230825/semi_static_layer/20230825090255.dat")
    # semi_img = ins.conv_map_2_image(semi_dat, 749, 531)
    # cv2.imshow("Image", semi_img)
    # cv2.waitKey()
    #
    # ex_dat = ins.map_read("/var/rdr/web/map/20230825/exclusion_zone_layer/20230825090253.dat")
    # ex_img = ins.conv_map_2_image(ex_dat, 749, 531)
    # cv2.imshow("Image", ex_img)
    # cv2.waitKey()
    #
    # env_dat = ins.map_read("/var/rdr/web/map/20230825/environment/20230825090256.dat")
    # env_img = ins.conv_map_2_image(env_dat, 749, 531)
    # cv2.imshow("Image", env_img)
    # cv2.waitKey()
    #