#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
"""
 Notification of regular map information

 @author East Japan Accounting Center Co., Ltd

 (c)2022 The University of Aizu
 This software is released under the New BSD License.
"""
import logging
from rdr_mqtt_pub import RdrMqttPub
from mongodb import MongoDB
from postgres import PostgreSQL

from map_type_info import MapLayerName


logger = logging.getLogger(__name__)


class EnvMapInfoNotice(RdrMqttPub):

    def __init__(self):
        super(EnvMapInfoNotice, self).__init__()
        self.mongodb = None
        self.mqtt_client = None
        self.postgres = None
        self.mongodb_conf = None
        self.postgres_conf = None
        self.db_conf_path = "./conf/db_conf.json"
        self.topic_name = "cmd/rdr/MapUpdateNotify"

    def connect_mongodb(self):
        """
        connect mongodb

        Return
        ----------
        result: connection result
        """
        result = False
        host = self.mongodb_conf["host"]
        port = self.mongodb_conf["port"]
        conn = self.mongodb.connection(host, port)
        if conn is not None:
            result = True
        else:
            logger.error("Unable to connect to MongoDB.")
        return result

    def disconnect_mongodb(self):
        """
        disconnect mongodb
        """
        self.mongodb.close()

    def find_psql(self, param):
        """
        Provide latest data for specified parameters from postgresql

        Param
        ----------
        param: column to search

        Return
        ----------
        find_data: search result data
        """
        find_data = None
        try:
            # connection postgres
            with self.postgres.get_connection() as conn:
                # get cursor
                with self.postgres.get_acquisition_cursor(conn) as cur:
                    request_sql = "SELECT * FROM {} ORDER BY {} DESC LIMIT 1"
                    # Search the latest environmental maps
                    find_res, find_data = self.postgres.select(cur, request_sql, param)
                    if not find_res or not find_data:
                        find_data = None
                    else:
                        # extract map data from list format
                        find_data = find_data.pop()
        except Exception as e:
            logger.error(e)
        return find_data

    def find_map_mongodb(self, query):
        """
        retrieve map layer from DB

        Param
        ----------
        map_layer: layer type of map to search

        Return
        ----------
        doc: search results
        """
        doc = self.mongodb.get_document(self.mongodb_conf["db_name"], self.mongodb_conf["map_collection"], query)
        return doc

    def find_layer_map_mongodb(self, env_map_info):
        """
        retrieve map layer list from DB

        Param
        ----------
        env_map_info:  of map to search

        Return
        ----------
        doc: search results
        """

        find_result = False
        map_layer_info_a = []
        if env_map_info is None:
            # Set the search condition to the latest layer map
            request_query_a = [{"map_layer": MapLayerName.STATIC.value, "space": "real"},
                               {"map_layer": MapLayerName.SEMI_STATIC.value, "space": "real"},
                               {"map_layer": MapLayerName.EXCLUSION_ZONE.value, "space": "real"},
                               {"map_layer": MapLayerName.SOCIO.value}]
        else:
            # Set the revision of the layer map associated with
            # the environment map in the search condition
            request_query_a = [{"revision": env_map_info["static_layer_revision"], "space": "real"},
                               {"revision": env_map_info["semi_static_layer_revision"], "space": "real"},
                               {"revision": env_map_info["exclusion_zone_layer_revision"], "space": "real"},
                               {"map_layer": MapLayerName.SOCIO.value},
                               {"env_revision": env_map_info["revision"]}]
        request_query_a += [{"map_layer": MapLayerName.STATIC.value, "space": "virtual"},
                            {"map_layer": MapLayerName.SEMI_STATIC.value, "space": "virtual"},
                            {"map_layer": MapLayerName.EXCLUSION_ZONE.value, "space": "virtual"},
                            {"map_layer": MapLayerName.ENVIRONMENT.value, "space": "virtual"}]

        # connection mongodb
        if self.connect_mongodb():
            # Search layer maps and store results as array data
            map_layer_info_a = [self.find_map_mongodb(query) for query in request_query_a]
            self.disconnect_mongodb()
            # Remove None from search results list
            map_layer_info_a = list(filter(lambda map_layer: map_layer is not None, map_layer_info_a))
            if len(map_layer_info_a) > 0:
                find_result = True
        return find_result, map_layer_info_a

    def config_setting(self):
        """
        preparation of configuration data

        Return
        ----------
        result: end of preparation
        """
        result = False
        parse_res, db_conf = self.is_json_file_parse(self.db_conf_path)
        if parse_res:
            self.mongodb_conf = db_conf["mongodb"]
            self.postgres_conf = db_conf["postgresql"]
            self.mongodb = MongoDB(self.mongodb_conf["host"], self.mongodb_conf["port"], self.mongodb_conf["db_name"])
            self.postgres = PostgreSQL(self.postgres_conf["db_name"], self.postgres_conf["host"],
                                       self.postgres_conf["tbl_name"], self.postgres_conf["user"],
                                       self.postgres_conf["password"])
            result = True
        return result

    @staticmethod
    def _set_map_info_dic(robot_id: str, space: str, time: str, revision: str, location_info: dict, map_layer: str):
        """
        create a format for dictionary data for notifications

        Param
        ----------
        robot_id: robot id
        time: time information
        revision: map revision
        location_info: location information
        map_layer: map layer type

        Return
        ----------
        dic: dictionary data for notifications
        """
        dic = {"id": robot_id, "space": space, "time": time, "revision": revision,
               "location_info": location_info,
               "map_layer": map_layer}
        return dic

    def _create_info_message(self, layer_map_info_a, env_map_info):
        """
        create message for map information notification

        Param
        ----------
        layer_map_info_a: layer map information list
        env_map_info: environment map information

        Return
        ----------
        return_message: create message
        """

        set_keys_a = ["id", "space", "time", "revision",  "location_info", "map_layer"]
        keys_a = ["robot_id", "space", "time", "revision", "location_info", "map_layer"]
        message_dic_a = []
        # layer map
        for layer_map_info in layer_map_info_a:
            if layer_map_info is not None:
                # Form dictionary format map data and store in array
                dic = {}
                # store the value corresponding to the key
                set_value_a = [layer_map_info.get(key) for key in keys_a]
                # Store key and value at the same time and update dictionary data
                dic.update(zip(set_keys_a, set_value_a))
                message_dic_a.append(dic)
        # environmental map
        if env_map_info is not None:
            set_id = env_map_info["robot_id"]
            set_space = env_map_info["space"]
            set_revision = env_map_info["revision"]
            time_str = "{}+0900".format(env_map_info["time"].strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3])
            location_info = {"location": env_map_info["location"], "lat": env_map_info["lat"],
                             "lon": env_map_info["lon"], "azimuth": env_map_info["azimuth"]}
            dic = self._set_map_info_dic(set_id, set_space, time_str, set_revision, location_info, MapLayerName.ENVIRONMENT.value)
            message_dic_a.append(dic)
        return message_dic_a

    def publish_map_info(self, layer_map_info_a, env_map_info):
        """
        publish map information

        Param
        ----------
        layer_map_info_a: layer map information list
        env_map_info: environment map information

        Return
        ----------
        result: publish result
        """
        # Create map data in send format
        send_msg = self._create_info_message(layer_map_info_a, env_map_info)
        result = False
        # Connect to mqtt and publish map information
        if self.setting(self.topic_name):
            if self.run(send_msg):
                result = True
            self.disconnect()
        else:
            logger.error("Unable to connect to MQTT Broker.")
        return result


def env_map_info_notice_proc(_option1, _option2, _option3, _option4):
    """
    map information notification process

    Param
    ----------
    _option1: fixed arguments
    _option2: fixed arguments
    _option3: fixed arguments
    _option4: fixed arguments

    Return
    ----------
    result: process result
    """
    result = False
    logger.debug("Map notification process started")
    map_notice_ins = EnvMapInfoNotice()
    if not map_notice_ins.config_setting():
        return result
    # find environmental map
    env_map_info = map_notice_ins.find_psql("time")
    # find layer map
    find_mongo_res, map_layer_info_a = map_notice_ins.find_layer_map_mongodb(env_map_info)
    if not find_mongo_res:
        logger.error("Unable to get the document.")
        return result
    # publish map data
    if map_notice_ins.publish_map_info(map_layer_info_a, env_map_info):
        result = True
        logger.debug("Notification completed")
    return result


def json_file_parse(path):
    """
    Connect to specified MQTT Broker.

    Param
    ----------
    path: json file path

    Return
    -------
    Parse result data
    """
    return EnvMapInfoNotice.is_json_file_parse(path)


if __name__ == "__main__":
    env_map_info_notice_proc(0, 0, 0, 0)
    # pass
    # res = env_map_info_notice_proc(0, 0, 0, 0)
    # ins = EnvMapInfoNotice()
    # res = ins.config_setting()
    # res = ins.connect_mongodb()
    # res = ins.find_map_mongodb({"map_layer": "socio"})

    # env_map_info = {"static_layer_revision": "1234-5678-abcd-efgh-1",
    #                 "semi_static_layer_revision": "1234-5678-abcd-abcd-1",
    #                 "exclusion_zone_layer_revision": "*-1",
    #                 "revision": "*-1"
    #                 }
    # env_map_info = None
    # res, map_layer_info_a = ins.find_layer_map_mongodb(env_map_info)
    #
    #
    # set_id = "robot"
    # time_str = "2022-08-08T23:59:59.000+09:00"
    # set_revision = "xxxx-xxxx-xxxx-xxxx-00"
    # location_info = {"location": "location", "lat": 255, "lon": 255, "azimuth": -1}
    # from datetime import datetime, timedelta, timezone
    # # res = ins._set_map_info_dic(set_id, time_str, set_revision, location_info, "map")
    # info_a = [{"revision":"1234-5678-abcd-efgh-1",
    #             "robot_id":"megarover_01",
    #             "time": "{}+0900".format(datetime.now(timezone(timedelta(hours=+9), 'JST')).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3]),
    #             "map_layer":"static_layer",
    #             "location_info":{"location":"lictia_1f","lat":0,"lon":0,"azimuth":0}}]

    # env_map_info = {"robot_id": "robot",
    #                 "time": "{}+0900".format(datetime.now(timezone(timedelta(hours=+9), 'JST')).strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]),
    #                 "revision": "xxxx-xxxx-xxxx-xxxx-01",
    #                 "location": "location",
    #                 "lat": 255, "lon": 255, "azimuth": -1}

    # env_map_info = ins.find_psql("time")
    # # info_a = [None]
    # # env_map_info = None
    #
    # res = ins.publish_map_info(info_a, env_map_info)

    # print(res)
