#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
"""
 Connect to and Export from DB

 @author East Japan Accounting Center Co., Ltd

 (c)2019 The University of Aizu
 This software is released under the New BSD License.
"""
import logging
import psycopg2
import psycopg2.extras


logger = logging.getLogger(__name__)

# File R/W
FILE_MODE_WRITE = "w"
TABLE_SEARCH_QUERY = "SELECT relname FROM pg_class WHERE relkind = 'r' AND relname =%s"


class PostgreSQL:
    """
    PostgreSQL operation(Connection, File export, ...)
    """

    def __init__(self, dbname, host, tbl_name, db_user, db_password):
        self._db_name = dbname
        self._tbl_name = tbl_name
        self._db_user = db_user
        self._db_password = db_password
        self._host = host
        self._db_client = None
        self._db_cursor = None

    def get_connection(self):
        """
        Connect to specified DB

        Returns
        -------
        DB instance
        """
        self._db_client = None

        try:
            doc = "dbname={} host={} user={} password={}".format(self._db_name, self._host, self._db_user, self._db_password)
            # self._db_client = psycopg2.connect(dbname=self._db_name, user=self._db_user, password=self._db_password)
            self._db_client = psycopg2.connect(doc)
        except psycopg2.Error as e:
            logger.error("Unable to connect to PostgreSQL.")
            logger.error(e)
        return self._db_client

    def get_acquisition_cursor(self, conn):
        """
        acquisition a cursor instance from a destination

        Parameters
        ----------
        conn: connection object
            DB connection instance

        Returns
        -------
        cursor instance
        """
        self._db_cursor = None
        if conn is not None:
            try:
                self._db_cursor = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            except psycopg2.Error as e:
                logger.error(e)

        return self._db_cursor

    def select(self, cur, sql, parameter):
        """
        Data select

        Parameters
        ----------
        cur: cursor
            cursor instance
        sql: string
            sql comand
        parameter: string
            select parameters

        Returns
        -------
        rtn: boolean
            select result
            True: Success
            False: Failure
        dict_data_list: list
            find fetch data
        """
        rtn = False
        dict_data_list = []
        # Validating table
        if not self._table_check(cur):
            return rtn, dict_data_list
        sql = sql.format(self._tbl_name, parameter)
        try:
            cur.execute(sql,)
        except psycopg2.Error as e:
            logger.error(e)
            logger.error("Unable to get the document.")
        else:
            if cur.rowcount > 0:
                dict_data_list = [dict(fetch_data) for fetch_data in cur.fetchall()]
            rtn = True
        return rtn, dict_data_list

    def insert(self, cur, sql, parameter):
        """
        Data insert

        Parameters
        ----------
        cur: cursor
            cursor instance
        sql: string
            sql comand
        parameter: string
            insert parameters

        Returns
        -------
        rtn: boolean
            insert result
            True: Success
            False: Failure
        """
        rtn = False
        # Validating table
        if not self._table_check(cur):
            return rtn
        sql = sql.format(self._tbl_name)
        try:
            cur.execute(sql, parameter)
        except psycopg2.Error as e:
            logger.error(e)
            logger.error("Unable to write the document.")
        else:
            if cur.rowcount > 0:
                rtn = True
        return rtn

    def _table_check(self, cur):
        """
        Validating table names

        Parameters
        ----------
        cur: cursor object
            DB cursor instance

        Returns
        -------
        rtn: boolean
            Export result
            True: Success
            False: Failure
        """
        rtn = False
        if cur is not None:
            try:
                cur.execute(TABLE_SEARCH_QUERY, (self._tbl_name,))
                if cur.fetchone():
                    rtn = True
            except psycopg2.Error as e:
                logger.error(e)

        return rtn


if __name__ == '__main__':
    pass
    # opt = { "db_name": "rdr_2nd",
    #         "host": "localhost",
    #         # "host": "local",
    #         "tbl_name": "env_map_tbl",
    #         "user": "uoa_rtcloud",
    #         "password": "UoAEACRTCloud2nd"
    #     }
    #
    # request_sql = "SELECT (left(max(revision), 20) || " \
    #               "cast(max(cast(substring(revision, 21, length(revision)) as integer)) " \
    #               "as character varying)) as revision " \
    #               "FROM {} WHERE revision LIKE  {}"
    # param = "'_____-____-____-____-%'"
    # request_sql = "SELECT * from {} where {}"
    # param = "type = \'robot2\'"
    #
    # request_sql = "INSERT INTO {} (id) VALUES (%s)"
    # param = (1, )
    #
    # ins = PostgreSQL(opt["db_name"], opt["host"], opt["tbl_name"], opt["user"], opt["password"])
    # try:
    #     with ins.get_connection() as conn:
    #         with ins.get_acquisition_cursor(conn) as cur:
    #             select_res = None
    #             select_data = None
    #             # select_res, select_data = ins.select(cur, request_sql, param)
    #             select_res = ins.insert(cur, request_sql, param)
    #
    #             print("{}, {}".format(select_res, select_data))
    # except Exception as e:
    #     logger.error(e)
