#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-
"""
 Processing that performs map-related management

 @author East Japan Accounting Center Co., Ltd

 (c)2022 The University of Aizu
 This software is released under the New BSD License.
"""
import json
import signal
import logging
import os
from datetime import datetime, timedelta, timezone
from enum import IntFlag
from concurrent.futures import ThreadPoolExecutor
from socket import socket, AF_INET, SOCK_STREAM, SHUT_RDWR
from env_map_info_notice import json_file_parse
from env_map_info_notice import env_map_info_notice_proc
from env_map_builder import env_map_builder_proc


log_dir = "./log"
if os.path.isdir(log_dir) is False:
    os.makedirs(log_dir)

if __name__ == "__main__":
    log_format = "%(asctime)s %(levelname)s %(name)s :%(message)s"
    log_file = f"{log_dir}/{datetime.now(timezone(timedelta(hours=+9), 'JST')).strftime('%Y%m%d%H%M%S')}.log"

    # stream handler setting
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(log_format))

    # file handler setting
    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(logging.Formatter(log_format))
    logging.basicConfig(level=logging.NOTSET, handlers=[stream_handler, file_handler])

logger = logging.getLogger(__name__)


class ReqMapProcId(IntFlag):
    MAP_INFO_NOTICE = 0
    ENV_MAP_BUILDER = 1
    PROC_MAX = 2


class ReqMapSocket:
    host = None
    port = None

    def __init__(self):
        self.max_message = 2048
        self.num_thread = 4
        self.server_socket = None

    def _send_message(self, dic):
        """
        send message with socket communication protocol
        Param
        ----------
        dic: dictionary data to send
        """
        try:
            if self.host is not None and self.port is not None:
                # socket connect
                send_sock = socket(AF_INET, SOCK_STREAM)
                send_sock.connect((self.host, self.port))
                # send message
                message = json.dumps(dic)
                send_sock.send(message.encode("utf-8"))
                # socket close
                send_sock.close()
            else:
                logger.error("Unable to connect to socket port.")

        except Exception as exc_e:
            logger.error(exc_e)
            logger.error("Unable to send socket message.")

    def set_socket_setting(self, host, port):
        self.__class__.host = host
        self.__class__.port = port

    def connection_server(self, host, port):
        """
        server side socket communication protocol connection

        Param
        ----------
        host: connection host name
        port: connection port number

        Return
        ----------
        result: connection result
        """
        result = False
        try:
            if host is not None and port is not None:
                self.__class__.host = host
                self.__class__.port = port
                self.server_socket = None
                self.server_socket = socket(AF_INET, SOCK_STREAM)
                self.server_socket.bind((self.host, self.port))
                self.server_socket.listen(self.num_thread)
                result = True
            else:
                logger.error("Unable to connect to socket port.")
        except Exception as e:
            logger.error(e)
            logger.error("Unable to connect to socket port.")
        return result

    def receive_server(self):
        """
        server-side socket communication message reception

        Return
        ----------
        receive_message: received message
        """
        receive_message = ""
        if self.server_socket is not None:
            conn, _ = self.server_socket.accept()
            receive_message = conn.recv(self.max_message)
            conn.close()
        return receive_message

    def close_server(self):
        """
        server side socket communication protocol disconnection
        """
        if self.server_socket is not None:
            self.server_socket.shutdown(SHUT_RDWR)
            self.server_socket.close()


class ReqMapSocketMsg(ReqMapSocket):
    def __init__(self, proc_id: ReqMapProcId, option1="", option2="", option3="", option4=""):
        super(ReqMapSocketMsg, self).__init__()
        self.message_dic = {"event_id": proc_id, "option1": option1,
                            "option2": option2, "option3": option3,
                            "option4": option4}

    def send(self):
        """
        send a socket message
        """
        self._send_message(self.message_dic)


class RdrMapManager:
    PERIODIC_NOTIFICATION_TIMER = 60

    def __init__(self):
        self.MAP_MGR_PRC_TBL = [
            env_map_info_notice_proc,  # MAP_INFO_NOTICE
            env_map_builder_proc  # ENV_MAP_BUILDER
            ]
        self.socket_conf_path = "./conf/socket_conf.json"

    @staticmethod
    def handler(_signum, _frame):
        """
        a signal handler that runs periodically

        Param
        ----------
         _signum: signal number
         _frame:  frame object
        """
        sock_ins = ReqMapSocketMsg(ReqMapProcId.MAP_INFO_NOTICE)
        sock_ins.send()

    @staticmethod
    def parse_message(message):
        """
        parse received socket message

        Param
        ----------
        message: received socket message

        Returns
        ----------
        result:  parse result
        event_id: requested event id
        option_a: requested option array
        """
        result = False
        event_id = None
        option_a = []
        try:
            dc_msg = message.decode(encoding="utf-8")
            request_msg = json.loads(dc_msg)
            event_id = request_msg["event_id"]
            option_a.append(request_msg["option1"])
            option_a.append(request_msg["option2"])
            option_a.append(request_msg["option3"])
            option_a.append(request_msg["option4"])
            if event_id is not None and event_id < ReqMapProcId.PROC_MAX:
                result = True
        except json.JSONDecodeError as json_e:
            logger.error(json_e)
        except Exception as other_e:
            logger.error(other_e)
        return result, event_id, option_a

    def main(self):
        """
        main processing of map manager.
        parse the received socket message and execute the corresponding process
        """
        logger.info("RDR Map Manager start")

        # config file parse
        parse_res, conf_dic = json_file_parse(self.socket_conf_path)
        if not parse_res:
            return  # Unable to read setting file.
        # Establishment of Socket communication
        sock_ins = ReqMapSocket()
        sock_res = sock_ins.connection_server(conf_dic["host"], conf_dic["port"])
        if not sock_res:
            return  # Unable to connect to socket port.
        # Register signal handler for signal.SIGALRM
        signal.signal(signal.SIGALRM, self.handler)
        # Signal.SIGALRM interrupt reservation
        signal.setitimer(signal.ITIMER_REAL,
                         RdrMapManager.PERIODIC_NOTIFICATION_TIMER,
                         RdrMapManager.PERIODIC_NOTIFICATION_TIMER)
        logger.debug("socket receive started")
        # Received message wait loop
        with ThreadPoolExecutor(max_workers=5, thread_name_prefix="thread") as pool:
            while True:
                try:
                    rcv_message = sock_ins.receive_server()
                    # socket message parse
                    req_message = self.parse_message(rcv_message)
                    parse_res, request_id = req_message[0:2]
                    if not parse_res:
                        continue
                    proc = self.MAP_MGR_PRC_TBL[request_id]
                    send_option_a = req_message[2]
                    logger.info("Thread[{}] started".format(request_id))
                    pool.submit(proc, *send_option_a)
                    logger.info("Thread[{}] end.".format(request_id))
                except KeyboardInterrupt:
                    logger.info("Detect interrupt key")
                    break
                except Exception as e:
                    logger.error(e)
        # signal handler end
        signal.alarm(0)
        # socket disconnect
        sock_ins.close_server()
        logger.info("RDR Map Manager End")
        return


if __name__ == "__main__":
    rdr_map_manager_ins = RdrMapManager()
    rdr_map_manager_ins.main()

