#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 Connect to the MQTT broker and various setting

 @author East Japan Accounting Center Co., Ltd

 (c)2021 The University of Aizu
 This software is released under the New BSD License.
"""
import json
import logging
import time

import paho.mqtt.client as mqtt

logger = logging.getLogger(__name__)


class RdrMqttPub(mqtt.Client):

    def __init__(self):
        super(RdrMqttPub, self).__init__()
        self._mqtt_broker_connecting = False
        self.CONNECT_RETRY_CNT = 3
        self.PUBLISH_WAIT = 1
        self.RETRY_PUBLISH_MAX = 3
        self._pub_topic = None

    def on_connect(self, mqttc, obj, flags, rc):
        """
        Called when the broker responds to our connection request.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        mqttc
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        flags
            response flags sent by the broker
        rc
            the connection result
            0: Connection successful
            1: Connection refused - incorrect protocol version
            2: Connection refused - invalid client identifier
            3: Connection refused - server unavailable
            4: Connection refused - bad username or password
            5: Connection refused - not authorised
            6-255: Currently unused.

        Returns
        -------
        None
        """
        logger.info("Connected with result code %s", str(rc))
        if rc == 0:
            self._mqtt_broker_connecting = True

    def on_publish(self, mqttc, obj, mid):
        """
        Called when the broker responds to a publish request.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        mqttc
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        mid
            See above URL

        Returns
        -------
        None
        """
        logger.info("Published: " + str(mid))

    def on_disconnect(self, mqttc, obj, rc):
        """
        Called when the client disconnects from the broker.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        mqttc
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        rc
            the connection result
            See detail on_connect()

        Returns
        -------
        None
        """
        logger.info("Disconnected. Result code %s", str(rc))
        self.loop_stop()
        self._mqtt_broker_connecting = False

    def _mqtt_broker_connect(self):
        """
        Connect to specified MQTT Broker.

        Returns
        -------
        Connect result
        True: Success False: Failure
        """

        result = False
        retry_count = 0

        parse_result, param = self.is_json_file_parse("./conf/mqtt_conf.json")
        if parse_result is False:
            return result

        try:
            # self.tls_set(ca_certs=param["ca_certs"], certfile=param["certfile"], keyfile=param["keyfile"], cert_reqs=mqtt.ssl.CERT_REQUIRED, tls_version=mqtt.ssl.PROTOCOL_TLSv1_2, ciphers=None)
            while not self._mqtt_broker_connecting and retry_count < self.CONNECT_RETRY_CNT:
                self.connect(param["host"], param["port"])
                self.loop_start()
                time.sleep(1)
                if self._mqtt_broker_connecting:
                    result = True
                    break
                else:
                    retry_count += 1
                    logger.info("MQTT Broker now connecting...")
        except Exception as e:
            logger.error(e)
        return result

    @staticmethod
    def is_json_file_parse(path):
        """
        Connect to specified MQTT Broker.

        Parameters
        ----------
        path: str
            json file path

        Returns
        -------
        Parse result
        True: Success False: Failure
        """

        result = False
        json_dict = {}

        try:
            with open(path) as f:
                json_dict = json.load(f)
                if "" not in json_dict.values():
                    result = True
                else:
                    logger.error("Param is empty.Please confirm.")
        except IOError:
            logger.error("Unable to read setting file.")
        except json.JSONDecodeError:
            logger.error("Unable to read setting file.")
        return result, json_dict

    def _publish(self, message):
        """
        Publish a message to the specified MQTT broker.
        Wait until receive the publish response.

        Parameters
        ----------
        message
            Publish message

        Returns
        ----------
        Publish result
        True: Success False: Failure
        """
        result = False
        for retry_count in range(self.RETRY_PUBLISH_MAX):
            # Publish message
            res = self.publish(self._pub_topic, message)
            # Result judgment
            if res[0] == mqtt.MQTT_ERR_SUCCESS:
                result = True
                break
            else:
                time.sleep(self.PUBLISH_WAIT)
        return result

    def setting(self, topic):
        """
        Connect to specified MQTT Broker.
        Wait until receive the specific topic

        Parameters
        ----------
        topic
            Specify the publish topic

        Returns
        -------
        None
        """
        result = False

        if self._mqtt_broker_connect():
            self._pub_topic = topic
            result = True

        return result

    def run(self, message):
        """
        Publish a message to specified MQTT Broker.

        Parameters
        ----------
        message
            publish message

        Returns
        -------
        None
        """
        result = False
        # Convert message to Json format
        cnv_message = json.dumps(message)

        if self._publish(cnv_message):
            result = True
        return result


# if __name__ == "__main__":
#     pass
    # mqttc = RdrMqttPub()
    # mqttc.run("obj_recognition_01/#")
